#include "utils.h"

void print_vector2f (vector2f v) {
  printf ("(%f, %f)", v.x, v.y);
}

void print_vector3f (vector3f v) {
  printf ("(%f, %f, %f)", v.x, v.y, v.z);
}

void print_vector4f (vector4f v) {
  printf ("(%f, %f, %f, %f)", v.x, v.y, v.z, v.t);
}

vector3f scale_vector3f (vector3f v, float s) {
  v.x *= s;
  v.y *= s;
  v.z *= s;
  return v;
}

vector4f scale_vector4f (vector4f v, float s) {
  v.x *= s;
  v.y *= s;
  v.z *= s;
  v.t *= s;
  return v;
}

float norm_vector3f (vector3f v) {
  return sqrt (v.x * v.x + v.y * v.y + v.z * v.z);
}

float norm_vector4f (vector4f v) {
  return sqrt (v.x * v.x + v.y * v.y + v.z * v.z + v.t * v.t);
}

vector3f normalize_vector3f (vector3f v) {
  return scale_vector3f (v, 1.0f / norm_vector3f (v));
}

vector4f normalize_vector4f (vector4f v) {
  return scale_vector4f (v, 1.0f / norm_vector4f (v));
}

float positive_fmod (float num, float denom) {
  float result = fmod (num, denom);
  return (result < 0.0f) ? result + denom : result;
}

const matrix mat_identity = { {
    1, 0, 0, 0,
    0, 1, 0, 0,
    0, 0, 1, 0,
    0, 0, 0, 1
  } };

matrix mat_prod (const matrix* m1, const matrix* m2) {
  matrix result;
  unsigned int row;
  unsigned int col;
  unsigned int ofs;

  for (row = 0; row < 4; row++) {
    ofs = 4 * row;
    for (col = 0; col < 4; col++) {
      result.coef[ofs + col] = (m1->coef[ofs] * m2->coef[col] +
				m1->coef[ofs + 1] * m2->coef[col + 4] +
				m1->coef[ofs + 2] * m2->coef[col + 8] +
				m1->coef[ofs + 3] * m2->coef[col + 12]);
    }
  }
  return result;
}

void mat_rotate_xt (matrix *m, float angle) {
  matrix r = mat_identity;
  float s = sin (angle);
  float c = cos (angle);
  r.coef[5] = c;
  r.coef[6] = -s;
  r.coef[9] = s;
  r.coef[10] = c;
  memcpy (m->coef, mat_prod (m, &r).coef, sizeof (m->coef));
}

void mat_rotate_yt (matrix *m, float angle) {
  matrix r = mat_identity;
  float s = sin (angle);
  float c = cos (angle);
  r.coef[0] = c;
  r.coef[8] = s;
  r.coef[2] = -s;
  r.coef[10] = c;
  memcpy (m->coef, mat_prod (m, &r).coef, sizeof (m->coef));
}

void mat_rotate_zt (matrix *m, float angle) {
  matrix r = mat_identity;
  float s = sin (angle);
  float c = cos (angle);
  r.coef[0] = c;
  r.coef[1] = -s;
  r.coef[4] = s;
  r.coef[5] = c;
  memcpy (m->coef, mat_prod (m, &r).coef, sizeof (m->coef));
}

void mat_rotate_xy (matrix *m, float angle) {
  matrix r = mat_identity;
  float s = sin (angle);
  float c = cos (angle);
  r.coef[10] = c;
  r.coef[11] = -s; //sign ???????????
  r.coef[14] = s;
  r.coef[15] = c;
  memcpy (m->coef, mat_prod (m, &r).coef, sizeof (m->coef));
}

matrix mat_proj (float fovy, float ar, float near, float far) {
  matrix result = { {0} };
  const float y_scale = 1.0f / tan (fovy / 2.0f);
  const float x_scale = y_scale / ar;
  const float len = far - near;

  result.coef[0] = x_scale;
  result.coef[5] = y_scale;
  result.coef[10] = - ((far + near) / len);
  result.coef[11] = -1;
  result.coef[14] = - ((2 * near * far) / len);
  return result;
}

float random_float () {
  return ((float) rand ()) / ((float) INT_MAX);
}

vector3f random_color () {
  vector3f result;
  result.x = random_float ();
  result.y = random_float ();
  result.z = random_float ();
  return result;
}

vector3f add_vector3f (vector3f v1, vector3f v2) {
  vector3f result;
  result.x = v1.x + v2.x;
  result.y = v1.y + v2.y;
  result.z = v1.z + v2.z;
  return result;
}
