#include "world.h"

void hopf_demo (world* w, int n_fibers, int n_vertices) {
  w->n_vertices = n_fibers * n_vertices;
  w->vertices = malloc (w->n_vertices * sizeof (vector4f));
  w->colors = malloc (w->n_vertices * sizeof (vector3f));

  w->n_faces = 2 * n_fibers * n_vertices;
  w->faces = malloc (w->n_faces * sizeof (GLuint));

  unsigned int fiber;
  unsigned int v;
  unsigned int ofs;
  vector3f vertex;
  vector3f color;
  for (fiber = 0; fiber < n_fibers; fiber++) {
    color = random_color ();
    for (v = 0; v < n_vertices; v++) {
      vertex.x = ((float) fiber) / ((float) n_fibers+1) * 2.0f * M_PI;
      vertex.y = ((float) fiber) / ((float) n_fibers+1) * 2.0f * M_PI;
      vertex.z = ((float) v) / ((float) n_vertices) * 2.0f * M_PI;

      ofs = fiber * n_vertices + v;
      w->vertices[ofs] = embedded_of_hopf (vertex);

      w->colors[ofs] = color;
      w->faces[2*ofs] = ofs;
      w->faces[2*ofs + 1] = (GLuint) (fiber * n_vertices + (v + 1) % n_vertices);
    }
  }
}

void init_vertices (world* w) {
  hopf_demo (w, 300, 300);
}

void init_buffers (world* w) {
  //boilerplate copypasta
  glGenVertexArrays (1, &w->vao);
  glBindVertexArray (w->vao);
  
  glGenBuffers (1, &w->vbo);
  glBindBuffer (GL_ARRAY_BUFFER, w->vbo);
  glBufferData (GL_ARRAY_BUFFER, w->n_vertices * sizeof (vector4f), w->vertices, GL_STATIC_DRAW);
  glVertexAttribPointer (0, 4, GL_FLOAT, GL_FALSE, 0, 0);
  glEnableVertexAttribArray (0);

  glGenBuffers (1, &w->color_buf);
  glBindBuffer (GL_ARRAY_BUFFER, w->color_buf);
  glBufferData (GL_ARRAY_BUFFER, w->n_vertices * sizeof (vector4f), w->colors, GL_STATIC_DRAW);
  glVertexAttribPointer (1, 3, GL_FLOAT, GL_FALSE, 0, 0);
  glEnableVertexAttribArray (1);

  glGenBuffers (1, &w->faces_buf);
  glBindBuffer (GL_ELEMENT_ARRAY_BUFFER, w->faces_buf);
  glBufferData (GL_ELEMENT_ARRAY_BUFFER, w->n_faces * sizeof (GLuint), w->faces, GL_STATIC_DRAW);

  exit_on_gl_failure ("Could not send world data to GPU.\n");
}

void init_shaders (world* w) {
  w->vertex_shader = load_shader ("main.vertex.glsl", GL_VERTEX_SHADER);
  w->fragment_shader = load_shader ("main.fragment.glsl", GL_FRAGMENT_SHADER);

  w->program_id = glCreateProgram ();
  glAttachShader (w->program_id, w->vertex_shader);
  glAttachShader (w->program_id, w->fragment_shader);
  glLinkProgram (w->program_id);

  exit_on_gl_failure ("Could not link shader.\n");

  w->rotation_matrix_loc = glGetUniformLocation (w->program_id, "rotation");
  w->projection_matrix_loc = glGetUniformLocation (w->program_id, "projection");

  exit_on_gl_failure ("Could not get uniform locations.\n");
}

world init_world () {
  world result;
  init_vertices (&result);
  init_buffers (&result);
  init_shaders (&result);
  return result;
}



void destroy_buffers (world w) {
  glDisableVertexAttribArray (1);
  glDisableVertexAttribArray (0);

  glBindBuffer (GL_ARRAY_BUFFER, 0);
  glDeleteBuffers (1, &w.color_buf);
  glDeleteBuffers (0, &w.vbo);

  glBindVertexArray (0);
  glDeleteVertexArrays (1, &w.vao);

  exit_on_gl_failure ("Could not destroy buffers.\n");
}

void destroy_shaders (world w) {
  glUseProgram (0);
  
  glDetachShader (w.program_id, w.fragment_shader);
  glDetachShader (w.program_id, w.vertex_shader);
  glDeleteShader (w.fragment_shader);
  glDeleteShader (w.vertex_shader);
  glDeleteProgram (w.program_id);

  exit_on_gl_failure ("Could not destroy shaders.\n");
}

void destroy_world (world w) {
  destroy_buffers (w);
  destroy_shaders (w);
  
  free (w.vertices);
  free (w.colors);
  free (w.faces);
}


void draw_world (world w, player p) {
  glUseProgram (w.program_id);
  glBindVertexArray (w.vao);
  //glBindBuffer (GL_ELEMENT_ARRAY_BUFFER, w.faces_buf);

  glUniformMatrix4fv (w.rotation_matrix_loc, 1, GL_FALSE, p.rot.coef);
  glUniformMatrix4fv (w.projection_matrix_loc, 1, GL_FALSE, p.proj.coef);

  //glDrawArrays (GL_POINTS, 0, w.n_faces);
  glDrawElements (GL_LINES, w.n_faces, GL_UNSIGNED_INT, 0);
  exit_on_gl_failure ("Could not draw.\n");

  //glBindBuffer (GL_ELEMENT_ARRAY_BUFFER, 0);
  glBindVertexArray (0);
  glUseProgram (0);
}
